"""Executes PAD pipeline"""


import logging

import click

from clapper.click import (
    ConfigCommand,
    ResourceOption,
    log_parameters,
    verbosity_option,
)

from bob.pipelines.distributed import (
    VALID_DASK_CLIENT_STRINGS,
    dask_get_partition_size,
)

logger = logging.getLogger(__name__)


@click.command(
    entry_point_group="bob.pad.config",
    cls=ConfigCommand,
    epilog="""\b
 Command line examples\n
 -----------------------


 $ bob pad run-pipeline my_experiment.py -vv
""",
)
@click.option(
    "--pipeline",
    "-p",
    required=True,
    entry_point_group="bob.pad.pipeline",
    help="Feature extraction algorithm",
    cls=ResourceOption,
)
@click.option(
    "--decision_function",
    "-f",
    show_default=True,
    default="decision_function",
    help="Name of the Pipeline step to call for results, eg. ``predict_proba``",
    cls=ResourceOption,
)
@click.option(
    "--database",
    "-d",
    required=True,
    entry_point_group="bob.pad.database",
    help="PAD Database connector (class that implements the methods: `fit_samples`, `predict_samples`)",
    cls=ResourceOption,
)
@click.option(
    "--dask-client",
    "-l",
    entry_point_group="dask.client",
    string_exceptions=VALID_DASK_CLIENT_STRINGS,
    default="single-threaded",
    help="Dask client for the execution of the pipeline.",
    cls=ResourceOption,
)
@click.option(
    "--group",
    "-g",
    "groups",
    type=click.Choice(["train", "dev", "eval"]),
    multiple=True,
    default=("dev", "eval"),
    help="If given, this value will limit the experiments belonging to a particular group",
    cls=ResourceOption,
)
@click.option(
    "-o",
    "--output",
    show_default=True,
    default="results",
    help="Saves scores (and checkpoints) in this folder.",
    cls=ResourceOption,
)
@click.option(
    "--checkpoint/--memory",
    "checkpoint",
    default=True,
    help="If --checkpoint (which is the default), all steps of the pipeline will be saved. Checkpoints will be saved in `--output`.",
    cls=ResourceOption,
)
@click.option(
    "--dask-partition-size",
    "-s",
    help="If using Dask, this option defines the size of each dask.bag.partition."
    "Use this option if the current heuristic that sets this value doesn't suit your experiment."
    "(https://docs.dask.org/en/latest/bag-api.html?highlight=partition_size#dask.bag.from_sequence).",
    default=None,
    type=click.INT,
    cls=ResourceOption,
)
@click.option(
    "--dask-n-workers",
    "-n",
    help="If using Dask, this option defines the number of workers to start your experiment."
    "Dask automatically scales up/down the number of workers due to the current load of tasks to be solved."
    "Use this option if the current amount of workers set to start an experiment doesn't suit you.",
    default=None,
    type=click.INT,
    cls=ResourceOption,
)
@click.option(
    "--no-dask",
    is_flag=True,
    help="If set, it will not use Dask for the execution of the pipeline.",
    cls=ResourceOption,
)
@verbosity_option(cls=ResourceOption, logger=logger)
def run_pipeline(
    pipeline,
    decision_function,
    database,
    dask_client,
    groups,
    output,
    checkpoint,
    dask_partition_size,
    dask_n_workers,
    no_dask,
    **kwargs,
):
    """Runs the simplest PAD pipeline."""

    log_parameters(logger)

    execute_pipeline(
        pipeline=pipeline,
        database=database,
        decision_function=decision_function,
        output=output,
        groups=groups,
        checkpoint=checkpoint,
        dask_client=dask_client,
        dask_partition_size=dask_partition_size,
        dask_n_workers=dask_n_workers,
        no_dask=no_dask,
    )


def execute_pipeline(
    pipeline,
    database,
    decision_function="decision_function",
    output="results",
    groups=("dev", "eval"),
    checkpoint=False,
    dask_client="single-threaded",
    dask_partition_size=None,
    dask_n_workers=None,
    no_dask=False,
):
    import os
    import sys

    import dask.bag

    import bob.pipelines as mario

    from bob.pipelines import DaskWrapper, is_pipeline_wrapped
    from bob.pipelines.distributed.sge import get_resource_requirements

    if no_dask:
        dask_client = None

    os.makedirs(output, exist_ok=True)

    if checkpoint:
        pipeline = mario.wrap(
            ["checkpoint"], pipeline, features_dir=output, model_path=output
        )

    # Fetching samples
    fit_samples = database.fit_samples()
    total_samples = len(fit_samples)
    predict_samples = dict()
    for group in groups:
        predict_samples[group] = database.predict_samples(group=group)
        total_samples += len(predict_samples[group])

    # Checking if the pipeline is dask-wrapped
    if (
        not any(is_pipeline_wrapped(pipeline, DaskWrapper))
    ) and dask_client is not None:
        # Scaling up if necessary
        if dask_n_workers is not None and not isinstance(dask_client, str):
            dask_client.cluster.scale(dask_n_workers)

        # Defining the partition size
        partition_size = None
        if not isinstance(dask_client, str):
            lower_bound = 1  # lower bound of 1 video per chunk since usually video are already big
            partition_size = dask_get_partition_size(
                dask_client.cluster, total_samples, lower_bound=lower_bound
            )
        if dask_partition_size is not None:
            partition_size = dask_partition_size

        pipeline = mario.wrap(["dask"], pipeline, partition_size=partition_size)

    # create an experiment info file
    with open(os.path.join(output, "Experiment_info.txt"), "wt") as f:
        f.write(f"{sys.argv!r}\n")
        f.write(f"database={database!r}\n")
        f.write("Pipeline steps:\n")
        for i, name, estimator in pipeline._iter():
            f.write(f"Step {i}: {name}\n{estimator!r}\n")

    # train the pipeline
    pipeline.fit(fit_samples)

    for group in groups:
        logger.info(f"Running PAD pipeline for group {group}")
        result = getattr(pipeline, decision_function)(predict_samples[group])

        resources = None
        if isinstance(result, dask.bag.core.Bag):
            resources = get_resource_requirements(pipeline)

        save_sample_scores(
            result=result,
            output=output,
            group=group,
            dask_client=dask_client,
            resources=resources,
        )

    logger.info("PAD experiment finished!")


def _get_csv_columns(sample):
    """Returns a dict of {csv_column_name: sample_attr_name} given a sample."""
    # Mandatory columns and their corresponding fields
    columns_attr = {
        "claimed_id": "subject",
        "test_label": "key",
        "is_bonafide": "is_bonafide",
        "attack_type": "attack_type",
        "score": "data",
    }
    # Preventing duplicates and unwanted data
    ignored_fields = list(columns_attr.values()) + ["annotations"]
    # Retrieving custom metadata attribute names
    metadata_fields = [
        k
        for k in sample.__dict__.keys()
        if not k.startswith("_") and k not in ignored_fields
    ]
    for field in metadata_fields:
        columns_attr[field] = field
    return columns_attr


def sample_to_dict_row(sample, columns_fields):
    row_values = {
        col: getattr(sample, attr, None) for col, attr in columns_fields.items()
    }
    return row_values


def score_samples_to_dataframe(samples):
    import pandas as pd

    rows, column_fields = [], None
    for sample in samples:
        if column_fields is None:
            column_fields = _get_csv_columns(sample)
        row_values = sample_to_dict_row(sample, column_fields)
        rows.append(row_values)
    df = pd.DataFrame(rows)
    return df


def save_sample_scores(
    result,
    output,
    group,
    dask_client,
    resources=None,
):
    import os

    import dask.bag
    import dask.dataframe as dd

    scores_path = os.path.join(output, f"scores-{group}.csv")

    if isinstance(result, dask.bag.core.Bag):
        # convert score samples to dataframes
        result = result.map_partitions(score_samples_to_dataframe)
        result = dd.from_delayed(result.to_delayed())
        result.to_csv(
            scores_path,
            single_file=True,
            compute_kwargs=dict(scheduler=dask_client, resources=resources),
            index=False,
        )

    else:
        # convert score samples to dataframes
        result = score_samples_to_dataframe(result)
        result.to_csv(scores_path, index=False)



"""Finalizes the scores that are produced by bob pad run-pipeline.
"""
import logging

import click

from clapper.click import log_parameters, verbosity_option

logger = logging.getLogger(__name__)


@click.command(
    name="finalize-scores",
    epilog="""\b
Examples:
  $ bob pad finalize_scores /path/to/scores-dev.csv
  $ bob pad finalize_scores /path/to/scores-{dev,eval}.csv
""",
)
@click.argument(
    "scores", type=click.Path(exists=True, dir_okay=False), nargs=-1
)
@click.option(
    "-m",
    "--method",
    default="mean",
    type=click.Choice(["mean", "median", "min", "max"]),
    show_default=True,
    help="The method to use when finalizing the scores.",
)
@click.option(
    "--backup/--no-backup", default=True, help="Whether to backup scores."
)
@verbosity_option(logger)
def finalize_scores(scores, method, backup, verbose):
    """Finalizes the scores given by bob pad run-pipeline
    When using bob.pad.base, Algorithms can produce several score values for
    each unique sample. You can use this script to average (or min/max) these
    scores to have one final score per sample.

    The conversion is done in-place (original files will be backed up).
    The order of scores will change.
    """
    import shutil

    import numpy
    import pandas as pd

    log_parameters(logger)

    mean = {
        "mean": numpy.nanmean,
        "median": numpy.nanmedian,
        "max": numpy.nanmax,
        "min": numpy.nanmin,
    }[method]

    for path in scores:
        logger.info("Finalizing scores in %s", path)

        if backup:
            logger.info("Backing up %s", path)
            shutil.copy(path, path + ".bak")

        df = pd.read_csv(path)

        # average the scores of each frame
        df["score"] = df.groupby("video_key")["score"].transform(mean)

        # remove frame_id column if it exists
        if "frame_id" in df.columns:
            df.drop("frame_id", axis=1, inplace=True)

        # make rows unique based on test_label
        df.drop_duplicates(subset=["video_key"], inplace=True)

        df.to_csv(path, index=False)



import os

import pkg_resources

from click.testing import CliRunner

from bob.io.base.testing_utils import assert_click_runner_result
from bob.pad.base.script import pad_commands


def test_gen_pad():
    dev_ref_file = pkg_resources.resource_filename(
        __name__, "data/csv_scores/scores-dev.csv"
    )
    eval_ref_file = pkg_resources.resource_filename(
        __name__, "data/csv_scores/scores-eval.csv"
    )
    with open(dev_ref_file, "r") as f:
        dev_ref = f.readlines()
    with open(eval_ref_file, "r") as f:
        eval_ref = f.readlines()
    runner = CliRunner()
    with runner.isolated_filesystem():
        cwd = "./"
        result = runner.invoke(
            pad_commands.gen,
            [
                cwd,
                "--mean-match",
                "10",
                "--mean-attacks",
                "9",
                "--mean-attacks",
                "6",
                "--n-attacks",
                "2",
            ],
        )
        assert_click_runner_result(result)
        with open(os.path.join(cwd, "scores-dev.csv"), "r") as f:
            for generated_line, reference in zip(f.readlines(), dev_ref):
                assert generated_line == reference
        with open(os.path.join(cwd, "scores-eval.csv"), "r") as f:
            for generated_line, reference in zip(f.readlines(), eval_ref):
                assert generated_line == reference


def test_det_pad():
    dev = pkg_resources.resource_filename(
        __name__, "data/csv_scores/scores-dev.csv"
    )
    test = pkg_resources.resource_filename(
        __name__, "data/csv_scores/scores-eval.csv"
    )
    runner = CliRunner()
    with runner.isolated_filesystem():
        result = runner.invoke(
            pad_commands.det, ["-e", "--output", "DET.pdf", dev, test]
        )
        assert_click_runner_result(result)


def test_hist_pad():
    dev = pkg_resources.resource_filename(
        __name__, "data/csv_scores/scores-dev.csv"
    )
    test = pkg_resources.resource_filename(
        __name__, "data/csv_scores/scores-eval.csv"
    )
    runner = CliRunner()
    with runner.isolated_filesystem():
        result = runner.invoke(pad_commands.hist, [dev])
        assert_click_runner_result(result)

    with runner.isolated_filesystem():
        result = runner.invoke(
            pad_commands.hist,
            [
                "--criterion",
                "min-hter",
                "--output",
                "HISTO.pdf",
                "-b",
                "30,20",
                dev,
            ],
        )
        assert_click_runner_result(result)

    with runner.isolated_filesystem():
        result = runner.invoke(
            pad_commands.hist,
            [
                "-e",
                "--criterion",
                "eer",
                "--output",
                "HISTO.pdf",
                "-b",
                "30",
                dev,
                test,
            ],
        )
        assert_click_runner_result(result)


def test_metrics_pad():
    dev = pkg_resources.resource_filename(
        __name__, "data/csv_scores/scores-dev.csv"
    )
    test = pkg_resources.resource_filename(
        __name__, "data/csv_scores/scores-eval.csv"
    )
    runner = CliRunner()
    with runner.isolated_filesystem():
        result = runner.invoke(pad_commands.metrics, ["-e", dev, test, "-vvv"])
        assert_click_runner_result(result)


def test_evaluate_pad():
    dev = pkg_resources.resource_filename(
        __name__, "data/csv_scores/scores-dev.csv"
    )
    test = pkg_resources.resource_filename(
        __name__, "data/csv_scores/scores-eval.csv"
    )
    runner = CliRunner()
    with runner.isolated_filesystem():
        result = runner.invoke(pad_commands.evaluate, [dev, test])
        assert_click_runner_result(result)



import h5py
import numpy as np

from bob.io.base.testing_utils import datafile
from bob.pad.base.error_utils import (
    apcer_bpcer,
    calc_threshold,
    split_csv_pad,
    split_csv_pad_per_pai,
)

GENERATE_REFERENCES = False

scores_dev = datafile("per_pai_scores/scores-dev.csv", module=__name__)
scores_dev_reference_mask = datafile(
    "per_pai_scores/scores-dev-{i}.hdf5", module=__name__
)


def _dump_dict(f, d, name):
    f[f"{name}_len"] = len(d)
    for i, (k, v) in enumerate(d.items()):
        f[f"{name}_key_{i}"] = k
        f[f"{name}_value_{i}"] = v


def _read_dict(f, name):
    ret = dict()
    for i in range(int(np.array(f[f"{name}_len"]))):
        k = np.array(f[f"{name}_key_{i}"])[0].decode()
        v = np.array(f[f"{name}_value_{i}"])
        if v.size > 1:
            v = v.tolist()
        else:
            v = v[0]
            if isinstance(v, bytes):
                v = v.decode()
        ret[k] = v
    return ret


def test_per_pai_apcer():
    for i, regexps in enumerate(
        (None, ["x[0-2]", "x[3-4]"], ["x[1-2]", "x[3-4]"])
    ):
        try:
            pos, negs = split_csv_pad_per_pai(scores_dev, regexps)
        except ValueError:
            if i == 2:
                continue
            raise
        all_negs = [s for scores in negs.values() for s in scores]

        thresholds = dict()
        for method in ("bpcer20", "far", "eer", "min-hter"):
            thresholds[method] = calc_threshold(
                method, pos, negs.values(), all_negs, far_value=0.1
            )

        metrics = dict()
        for method, threshold in thresholds.items():
            apcers, apcer, bpcer = apcer_bpcer(threshold, pos, *negs.values())
            metrics[method] = apcers + [apcer, bpcer]

        scores_dev_reference = scores_dev_reference_mask.format(i=i)
        if GENERATE_REFERENCES:
            with h5py.File(scores_dev_reference, "w") as f:
                f["pos"] = pos
                _dump_dict(f, negs, "negs")
                _dump_dict(f, thresholds, "thresholds")
                _dump_dict(f, metrics, "metrics")

        with h5py.File(scores_dev_reference, "r") as f:
            ref_pos = np.array(f["pos"]).tolist()
            ref_negs = _read_dict(f, "negs")
            ref_thresholds = _read_dict(f, "thresholds")
            ref_metrics = _read_dict(f, "metrics")

        assert pos == ref_pos
        assert negs == ref_negs
        assert thresholds == ref_thresholds
        assert metrics == ref_metrics


def test_csv_split():
    neg, pos = split_csv_pad(scores_dev)
    assert len(neg) == 5000, len(neg)
    assert len(pos) == 5000, len(pos)
    assert np.isclose(np.mean(neg), -10, atol=0.1), np.mean(neg)
    assert np.isclose(np.mean(pos), 10, atol=0.1), np.mean(pos)



from __future__ import annotations

from abc import ABCMeta, abstractmethod

from bob.pipelines import Sample


class Database(metaclass=ABCMeta):
    """Base database class for PAD experiments."""

    @abstractmethod
    def fit_samples(self) -> list[Sample]:
        """Returns :any:`bob.pipelines.Sample`'s to train a PAD model.

        Returns
        -------
        samples : list
            List of samples for model training.
        """
        pass

    @abstractmethod
    def predict_samples(self, group: str = "dev") -> list[Sample]:
        """Returns :any:`bob.pipelines.Sample`'s to be scored.

        Parameters
        ----------
        group : :py:class:`str`, optional
            Limits samples to this group

        Returns
        -------
        samples : list
            List of samples to be scored.
        """
        pass

    def all_samples(
        self, groups: str | list[str] | None = None
    ) -> list[Sample]:
        """Returns all the samples of the database in one list.

        Giving ``groups`` will restrict the ``predict_samples`` to those groups.
        """
        samples = self.fit_samples()
        if groups is not None:
            if type(groups) is str:
                groups = [groups]
            for group in groups:
                samples.extend(self.predict_samples(group=group))
        else:
            samples.extend(self.predict_samples(group=group))
        return samples



"""Runs error analysis on score sets, outputs metrics and plots"""

import click
import numpy as np

from tabulate import tabulate

import bob.bio.base.script.figure as bio_figure
import bob.measure.script.figure as measure_figure

from bob.measure import f_score, farfrr, precision_recall, roc_auc_score
from bob.measure.utils import get_fta_list

from ..error_utils import apcer_bpcer, calc_threshold


def _normalize_input_scores(input_score, input_name):
    pos, negs = input_score
    # convert scores to sorted numpy arrays and keep a copy of all negatives
    pos = np.ascontiguousarray(pos)
    pos.sort()
    all_negs = np.ascontiguousarray([s for neg in negs.values() for s in neg])
    all_negs.sort()
    # FTA is calculated on pos and all_negs so we remove nans from negs
    for k, v in negs.items():
        v = np.ascontiguousarray(v)
        v.sort()
        negs[k] = v[~np.isnan(v)]
    neg_list, pos_list, fta_list = get_fta_list([(all_negs, pos)])
    all_negs, pos, fta = neg_list[0], pos_list[0], fta_list[0]
    return input_name, pos, negs, all_negs, fta


class Metrics(bio_figure.Metrics):
    """Compute metrics from score files"""

    def __init__(self, ctx, scores, evaluation, func_load, names):
        if isinstance(names, str):
            names = names.split(",")
        super(Metrics, self).__init__(ctx, scores, evaluation, func_load, names)

    def get_thres(self, criterion, pos, negs, all_negs, far_value):
        return calc_threshold(
            criterion,
            pos=pos,
            negs=negs.values(),
            all_negs=all_negs,
            far_value=far_value,
            is_sorted=True,
        )

    def _numbers(self, threshold, pos, negs, all_negs, fta):
        pais = list(negs.keys())
        apcer_pais, apcer_ap, bpcer = apcer_bpcer(
            threshold, pos, *[negs[k] for k in pais]
        )
        apcer_pais = {k: apcer_pais[i] for i, k in enumerate(pais)}
        acer = (apcer_ap + bpcer) / 2.0
        fpr, fnr = farfrr(all_negs, pos, threshold)
        hter = (fpr + fnr) / 2.0
        far = fpr * (1 - fta)
        frr = fta + fnr * (1 - fta)

        nn = all_negs.shape[0]  # number of attack
        fp = int(round(fpr * nn))  # number of false positives
        np = pos.shape[0]  # number of bonafide
        fn = int(round(fnr * np))  # number of false negatives

        # precision and recall
        precision, recall = precision_recall(all_negs, pos, threshold)

        # f_score
        f1_score = f_score(all_negs, pos, threshold, 1)

        # auc
        auc = roc_auc_score(all_negs, pos)
        auc_log = roc_auc_score(all_negs, pos, log_scale=True)

        metrics = dict(
            apcer_pais=apcer_pais,
            apcer_ap=apcer_ap,
            bpcer=bpcer,
            acer=acer,
            fta=fta,
            fpr=fpr,
            fnr=fnr,
            hter=hter,
            far=far,
            frr=frr,
            fp=fp,
            nn=nn,
            fn=fn,
            np=np,
            precision=precision,
            recall=recall,
            f1_score=f1_score,
            auc=auc,
        )
        metrics["auc-log-scale"] = auc_log
        return metrics

    def _strings(self, metrics):
        n_dec = ".%df" % self._decimal
        for k, v in metrics.items():
            if k in ("precision", "recall", "f1_score", "auc", "auc-log-scale"):
                metrics[k] = "%s" % format(v, n_dec)
            elif k in ("np", "nn", "fp", "fn"):
                continue
            elif k in ("fpr", "fnr"):
                if "fp" in metrics:
                    metrics[k] = "%s%% (%d/%d)" % (
                        format(100 * v, n_dec),
                        metrics["fp" if k == "fpr" else "fn"],
                        metrics["nn" if k == "fpr" else "np"],
                    )
                else:
                    metrics[k] = "%s%%" % format(100 * v, n_dec)
            elif k == "apcer_pais":
                metrics[k] = {
                    k1: "%s%%" % format(100 * v1, n_dec) for k1, v1 in v.items()
                }
            else:
                metrics[k] = "%s%%" % format(100 * v, n_dec)

        return metrics

    def _get_all_metrics(self, idx, input_scores, input_names):
        """Compute all metrics for dev and eval scores"""
        for i, (score, name) in enumerate(zip(input_scores, input_names)):
            input_scores[i] = _normalize_input_scores(score, name)

        dev_file, dev_pos, dev_negs, dev_all_negs, dev_fta = input_scores[0]
        if self._eval:
            (
                eval_file,
                eval_pos,
                eval_negs,
                eval_all_negs,
                eval_fta,
            ) = input_scores[1]

        threshold = (
            self.get_thres(
                self._criterion, dev_pos, dev_negs, dev_all_negs, self._far
            )
            if self._thres is None
            else self._thres[idx]
        )

        title = self._legends[idx] if self._legends is not None else None
        if self._thres is None:
            far_str = ""
            if self._criterion == "far" and self._far is not None:
                far_str = str(self._far)
            click.echo(
                "[Min. criterion: %s %s] Threshold on Development set `%s`: %e"
                % (
                    self._criterion.upper(),
                    far_str,
                    title or dev_file,
                    threshold,
                ),
                file=self.log_file,
            )
        else:
            click.echo(
                "[Min. criterion: user provided] Threshold on "
                "Development set `%s`: %e" % (dev_file or title, threshold),
                file=self.log_file,
            )

        res = []
        res.append(
            self._strings(
                self._numbers(
                    threshold, dev_pos, dev_negs, dev_all_negs, dev_fta
                )
            )
        )

        if self._eval:
            # computes statistics for the eval set based on the threshold a priori
            res.append(
                self._strings(
                    self._numbers(
                        threshold, eval_pos, eval_negs, eval_all_negs, eval_fta
                    )
                )
            )
        else:
            res.append(None)

        return res

    def compute(self, idx, input_scores, input_names):
        """Compute metrics for the given criteria"""
        title = self._legends[idx] if self._legends is not None else None
        all_metrics = self._get_all_metrics(idx, input_scores, input_names)
        headers = [" " or title, "Development"]
        if self._eval:
            headers.append("Evaluation")
        rows = []

        for name in self.names:
            if name == "apcer_pais":
                for k, v in all_metrics[0][name].items():
                    print_name = f"APCER ({k})"
                    rows += [[print_name, v]]
                    if self._eval:
                        rows[-1].append(all_metrics[1][name][k])
                continue
            print_name = name.upper()
            rows += [[print_name, all_metrics[0][name]]]
            if self._eval:
                rows[-1].append(all_metrics[1][name])

        click.echo(tabulate(rows, headers, self._tablefmt), file=self.log_file)


class MultiMetrics(Metrics):
    """Compute metrics from score files"""

    def __init__(self, ctx, scores, evaluation, func_load, names):
        super(MultiMetrics, self).__init__(
            ctx, scores, evaluation, func_load, names=names
        )
        self.rows = []
        self.headers = None
        self.pais = None

    def _compute_headers(self, pais):
        names = list(self.names)
        if "apcer_pais" in names:
            idx = names.index("apcer_pais")
            names = (
                [n.upper() for n in names[:idx]]
                + self.pais
                + [n.upper() for n in names[idx + 1 :]]
            )
        self.headers = ["Methods"] + names
        if self._eval and "hter" in self.names:
            self.headers.insert(1, "HTER (dev)")

    def _strings(self, metrics):
        formatted_metrics = dict()
        for name in self.names:
            if name == "apcer_pais":
                for pai in self.pais:
                    mean = metrics[pai].mean()
                    std = metrics[pai].std()
                    mean = super()._strings({pai: mean})[pai]
                    std = super()._strings({pai: std})[pai]
                    formatted_metrics[pai] = f"{mean} ({std})"
            else:
                mean = metrics[name].mean()
                std = metrics[name].std()
                mean = super()._strings({name: mean})[name]
                std = super()._strings({name: std})[name]
                formatted_metrics[name] = f"{mean} ({std})"

        return formatted_metrics

    def _structured_array(self, metrics):
        names = list(metrics[0].keys())
        if "apcer_pais" in names:
            idx = names.index("apcer_pais")
            pais = list(
                f"APCER ({pai})" for pai in metrics[0]["apcer_pais"].keys()
            )
            names = names[:idx] + pais + names[idx + 1 :]
            self.pais = self.pais or pais
        formats = [float] * len(names)
        dtype = dict(names=names, formats=formats)
        array = []
        for each in metrics:
            array.append([])
            for k, v in each.items():
                if k == "apcer_pais":
                    array[-1].extend(list(v.values()))
                else:
                    array[-1].append(v)
        array = [tuple(a) for a in array]
        return np.array(array, dtype=dtype)

    def compute(self, idx, input_scores, input_names):
        """Computes the average of metrics over several protocols."""
        for i, (score, name) in enumerate(zip(input_scores, input_names)):
            input_scores[i] = _normalize_input_scores(score, name)

        step = 2 if self._eval else 1
        self._dev_metrics = []
        self._thresholds = []
        for scores in input_scores[::step]:
            name, pos, negs, all_negs, fta = scores
            threshold = (
                self.get_thres(self._criterion, pos, negs, all_negs, self._far)
                if self._thres is None
                else self._thres[idx]
            )
            self._thresholds.append(threshold)
            self._dev_metrics.append(
                self._numbers(threshold, pos, negs, all_negs, fta)
            )
        self._dev_metrics = self._structured_array(self._dev_metrics)

        if self._eval:
            self._eval_metrics = []
            for i, scores in enumerate(input_scores[1::step]):
                name, pos, negs, all_negs, fta = scores
                threshold = self._thresholds[i]
                self._eval_metrics.append(
                    self._numbers(threshold, pos, negs, all_negs, fta)
                )
            self._eval_metrics = self._structured_array(self._eval_metrics)

        title = self._legends[idx] if self._legends is not None else name

        dev_metrics = self._strings(self._dev_metrics)

        if self._eval and "hter" in dev_metrics:
            self.rows.append([title, dev_metrics["hter"]])
        elif not self._eval:
            row = [title]
            for name in self.names:
                if name == "apcer_pais":
                    for pai in self.pais:
                        row += [dev_metrics[pai]]
                else:
                    row += [dev_metrics[name]]
            self.rows.append(row)
        else:
            self.rows.append([title])

        if self._eval:
            eval_metrics = self._strings(self._eval_metrics)
            row = []
            for name in self.names:
                if name == "apcer_pais":
                    for pai in self.pais:
                        row += [eval_metrics[pai]]
                else:
                    row += [eval_metrics[name]]

            self.rows[-1].extend(row)

        # compute header based on found PAI names
        if self.headers is None:
            self._compute_headers(self.pais)

    def end_process(self):
        click.echo(
            tabulate(self.rows, self.headers, self._tablefmt),
            file=self.log_file,
        )
        super(MultiMetrics, self).end_process()


class Roc(bio_figure.Roc):
    """ROC for PAD"""

    def __init__(self, ctx, scores, evaluation, func_load):
        super(Roc, self).__init__(ctx, scores, evaluation, func_load)
        self._x_label = ctx.meta.get("x_label") or "APCER"
        default_y_label = "1-BPCER" if self._tpr else "BPCER"
        self._y_label = ctx.meta.get("y_label") or default_y_label


class Det(bio_figure.Det):
    def __init__(self, ctx, scores, evaluation, func_load):
        super(Det, self).__init__(ctx, scores, evaluation, func_load)
        self._x_label = ctx.meta.get("x_label") or "APCER (%)"
        self._y_label = ctx.meta.get("y_label") or "BPCER (%)"


class Hist(measure_figure.Hist):
    """Histograms for PAD"""

    def _setup_hist(self, neg, pos):
        self._title_base = "PAD"
        self._density_hist(pos[0], n=0, label="Bona-fide", color="C1")
        self._density_hist(
            neg[0],
            n=1,
            label="Presentation attack",
            alpha=0.4,
            color="C7",
            hatch="\\\\",
        )



import glob
import os

import numpy as np

from click.testing import CliRunner
from sklearn.discriminant_analysis import LinearDiscriminantAnalysis
from sklearn.model_selection import train_test_split
from sklearn.pipeline import Pipeline

import bob.measure
import bob.pipelines as mario

from bob.io.base.testing_utils import assert_click_runner_result
from bob.pad.base.error_utils import split_csv_pad
from bob.pad.base.pipelines import Database
from bob.pad.base.script.run_pipeline import run_pipeline as run_pipeline_cli


def dataset_fixed_cov():
    # dummy data from sklearn
    # code from https://scikit-learn.org/stable/auto_examples/classification/plot_lda_qda.html
    """Generate 2 Gaussians samples with the same covariance matrix"""
    n, dim = 300, 2
    np.random.seed(0)
    C = np.array([[0.0, -0.23], [0.83, 0.23]])
    X = np.r_[
        np.dot(np.random.randn(n, dim), C),
        np.dot(np.random.randn(n, dim), C) + np.array([1, 1]),
    ]
    y = np.hstack((np.zeros(n), np.ones(n)))
    return X, y


class DummyPadDatabase(Database):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        X, y = dataset_fixed_cov()
        # split train and test
        X_train, X_test, y_train, y_test = train_test_split(
            X, y, test_size=0.33, random_state=42
        )
        self._fit_samples = [
            mario.Sample(x_, is_bonafide=bool(y_), key=str(i))
            for i, (x_, y_) in enumerate(zip(X_train, y_train))
        ]
        self._test_samples = [
            mario.Sample(x_, is_bonafide=bool(y_), key=str(i))
            for i, (x_, y_) in enumerate(zip(X_test, y_test))
        ]
        self._fit_samples = tuple(self._fit_samples)
        self._test_samples = tuple(self._test_samples)

    def fit_samples(self):
        return self._fit_samples

    def predict_samples(self, group="dev"):
        return self._test_samples


def dummy_pipeline():
    classifier = mario.wrap(
        ["sample"],
        LinearDiscriminantAnalysis(),
        fit_extra_arguments=[("y", "is_bonafide")],
    )
    pipeline = Pipeline([("lda", classifier)])
    return pipeline


def _create_config_file(path):
    with open(path, "w") as f:
        f.write(
            """
from tests.test_pipelines import DummyPadDatabase, dummy_pipeline
database = DummyPadDatabase()
pipeline = dummy_pipeline()
"""
        )


def test_run_pipeline():
    for options in [
        ["--no-dask", "--memory"],
        ["--no-dask"],
        ["--memory"],
        [],
    ]:
        runner = CliRunner()
        with runner.isolated_filesystem():
            _create_config_file("config.py")
            result = runner.invoke(
                run_pipeline_cli,
                [
                    "-vv",
                    "config.py",
                ]
                + options,
            )
            assert_click_runner_result(result)
            generated_files = glob.glob("results/**", recursive=True)

            # check if score file is generated
            assert os.path.isfile("results/scores-dev.csv"), generated_files
            assert os.path.isfile("results/scores-eval.csv"), generated_files

            # check if model is checkpointed if not --memory
            if "--memory" not in options:
                assert os.path.isfile("results/lda.pkl"), generated_files
            else:
                assert not os.path.isfile("results/lda.pkl"), generated_files

            # test if eer is accepable
            neg, pos = split_csv_pad("results/scores-dev.csv")
            eer = bob.measure.eer(neg, pos)
            assert eer < 0.08



#!/usr/bin/env python
# vim: set fileencoding=utf-8 :


from bob.bio.base.database.legacy import check_parameters_for_validity
from bob.pad.base.pipelines.abstract_classes import Database
from bob.pipelines.dataset import FileListDatabase


def validate_pad_sample(sample):
    if not hasattr(sample, "subject"):
        raise RuntimeError(
            "PAD samples should contain a `subject` attribute which "
            "reveals the identifies the person from whom the sample is created."
        )
    if not hasattr(sample, "attack_type"):
        raise RuntimeError(
            "PAD samples should contain a `attack_type` attribute which "
            "should be '' for bona fide samples and something like "
            "print, replay, mask, etc. for attacks. This attribute is "
            "considered the PAI type of each attack is used to compute APCER."
        )
    if sample.attack_type == "":
        sample.attack_type = None
    sample.is_bonafide = sample.attack_type is None
    if not hasattr(sample, "key"):
        sample.key = sample.filename
    return sample


class FileListPadDatabase(Database, FileListDatabase):
    """A PAD database interface from CSV files."""

    def __init__(
        self,
        name,
        dataset_protocols_path,
        protocol,
        transformer=None,
        **kwargs,
    ):
        super().__init__(
            name=name,
            dataset_protocols_path=dataset_protocols_path,
            protocol=protocol,
            transformer=transformer,
            **kwargs,
        )

    def __repr__(self) -> str:
        return "FileListPadDatabase(dataset_protocols_path='{}', protocol='{}', transformer={})".format(
            self.dataset_protocols_path, self.protocol, self.transformer
        )

    def purposes(self):
        return ("real", "attack")

    def samples(self, groups=None, purposes=None):
        results = super().samples(groups=groups)
        purposes = check_parameters_for_validity(
            purposes, "purposes", self.purposes(), self.purposes()
        )

        def _filter(s):
            return (s.is_bonafide and "real" in purposes) or (
                (not s.is_bonafide) and "attack" in purposes
            )

        results = [validate_pad_sample(sample) for sample in results]
        results = list(filter(_filter, results))
        return results

    def fit_samples(self):
        return self.samples(groups="train")

    def predict_samples(self, group="dev"):
        return self.samples(groups=group)



#!/usr/bin/env python
# Ivana Chingovska <ivana.chingovska@idiap.ch>
# Fri Dec  7 12:33:37 CET 2012
"""Utility functions for computation of EPSC curve and related measurement"""

import logging
import re

from collections import defaultdict

import numpy

from bob.bio.base.score.load import _iterate_csv_score_file
from bob.measure import (
    eer_threshold,
    far_threshold,
    farfrr,
    frr_threshold,
    min_hter_threshold,
)

logger = logging.getLogger(__name__)


def calc_threshold(
    method, pos, negs, all_negs, far_value=None, is_sorted=False
):
    """Calculates the threshold based on the given method.

    Parameters
    ----------
    method : str
        One of ``bpcer20``, ``eer``, ``min-hter``, ``apcer20``.
    pos : ``array_like``
        The positive scores. They should be sorted!
    negs : list
        A list of array_like negative scores. Each item in the list corresponds to
        scores of one PAI.
    all_negs : ``array_like``
        An array of all negative scores. This can be calculated from negs as well but we
        ask for it since you might have it already calculated.
    far_value : :obj:`float`, optional
        If method is far, far_value and all_negs are used to calculate the threshold.
    is_sorted : :obj:`bool`, optional
        If True, it means all scores are sorted and no sorting will happen.

    Returns
    -------
    float
        The calculated threshold.

    Raises
    ------
    ValueError
        If method is unknown.
    """
    method = method.lower()
    if "bpcer" in method:
        desired_apcer = 1 / float(method.replace("bpcer", ""))
        threshold = apcer_threshold(
            desired_apcer, pos, *negs, is_sorted=is_sorted
        )
    elif "apcer" in method:
        desired_bpcer = 1 / float(method.replace("apcer", ""))
        threshold = frr_threshold(
            all_negs, pos, desired_bpcer, is_sorted=is_sorted
        )
    elif method == "far":
        threshold = far_threshold(all_negs, pos, far_value, is_sorted=is_sorted)
    elif method == "eer":
        threshold = eer_threshold(all_negs, pos, is_sorted=is_sorted)
    elif method == "min-hter":
        threshold = min_hter_threshold(all_negs, pos, is_sorted=is_sorted)
    else:
        raise ValueError("Unknown threshold criteria: {}".format(method))

    return threshold


def apcer_threshold(desired_apcer, pos, *negs, is_sorted=False):
    """Computes the threshold given the desired APCER as the criteria.

    APCER is computed as max of all APCER_PAI values.
    The threshold will be computed such that the real APCER is **at most** the desired
    value.

    Parameters
    ----------
    desired_apcer : float
        The desired APCER value.
    pos : list
        An array or list of positive scores in float.
    *negs
        A list of negative scores. Each item corresponds to the negative scores of one
        PAI.
    is_sorted : :obj:`bool`, optional
        Set to ``True`` if ALL arrays (pos and negs) are sorted.

    Returns
    -------
    float
        The computed threshold that satisfies the desired APCER.
    """
    threshold = max(
        far_threshold(neg, pos, desired_apcer, is_sorted=is_sorted)
        for neg in negs
    )
    return threshold


def apcer_bpcer(threshold, pos, *negs):
    """Computes APCER_PAI, APCER, and BPCER given the positive scores and a list of
    negative scores and a threshold.

    Parameters
    ----------
    threshold : float
        The threshold to be used to compute the error rates.
    pos : list
        An array or list of positive scores in float.
    *negs
        A list of negative scores. Each item corresponds to the negative scores of one
        PAI.

    Returns
    -------
    tuple
        A tuple such as (list of APCER_PAI, APCER, BPCER)
    """
    apcers = []
    assert len(negs) > 0, negs
    for neg in negs:
        far, frr = farfrr(neg, pos, threshold)
        apcers.append(far)
    bpcer = frr  # bpcer will be the same in all cases
    return apcers, max(apcers), bpcer


def split_csv_pad_per_pai(filename, regexps=[], regexp_column="attack_type"):
    """Returns scores for Bona-Fide samples and scores for each PAI.
    By default, the real_id column (second column) is used as indication for
    each Presentation Attack Instrument (PAI).

    For example, with default regexps and regexp_column, if you have scores
    like::

        claimed_id, test_label,              is_bonafide, attack_type, score
        001,        bona_fide_sample_1_path, True,        ,            0.9
        001,        print_sample_1_path,     False,       print,       0.6
        001,        print_sample_2_path,     False,       print,       0.6
        001,        replay_sample_1_path,    False,       replay,      0.2
        001,        replay_sample_2_path,    False,       replay,      0.2
        001,        mask_sample_1_path,      False,       mask,        0.5
        001,        mask_sample_2_path,      False,       mask,        0.5

    this function will return 1 set of positive scores, and 3 sets of negative
    scores (for each print, replay, and mask PAIs).

    Otherwise, you can provide a list regular expressions that match each PAI.
    For example, with regexps as ['print', 'replay', 'mask'], if you have scores
    like::

        claimed_id, test_label,              is_bonafide, attack_type, score
        001,        bona_fide_sample_1_path, True,        ,            0.9
        001,        print_sample_1_path,     False,       print/1,     0.6
        001,        print_sample_2_path,     False,       print/2,     0.6
        001,        replay_sample_1_path,    False,       replay/1,    0.2
        001,        replay_sample_2_path,    False,       replay/2,    0.2
        001,        mask_sample_1_path,      False,       mask/1,      0.5
        001,        mask_sample_2_path,      False,       mask/2,      0.5

    the function will return 3 sets of negative scores (for print, replay, and
    mask PAIs, given in regexp).


    Parameters
    ----------
    filename : str
        Path to the score file.
    regexps : :obj:`list`, optional
        A list of regular expressions that match each PAI. If not given, the
        values in the column pointed by regexp_column are used to find scores
        for different PAIs.
    regexp_column : :obj:`str`, optional
        If a list of regular expressions are given, those patterns will be
        matched against the values in this column. default: ``attack_type``

    Returns
    -------
    tuple
        A tuple, ([positives], {'pai_name': [negatives]}), containing positive
        scores and a dict of negative scores mapping PAIs names to their
        respective scores.

    Raises
    ------
    ValueError
        If none of the given regular expressions match the values in
        regexp_column.
    KeyError
        If regexp_column is not a column of the CSV file.
    """
    pos = []
    negs = defaultdict(list)
    logger.debug(f"Loading CSV score file: '{filename}'")
    if regexps:
        regexps = [re.compile(pattern) for pattern in regexps]

    for row in _iterate_csv_score_file(filename):
        # if it is a Bona-Fide score
        if row["is_bonafide"].lower() == "true":
            pos.append(row["score"])
            continue
        if not regexps:
            negs[row[regexp_column]].append(row["score"])
            continue
        # if regexps is not None or empty and is not a Bona-Fide score
        for pattern in regexps:
            if pattern.search(row[regexp_column]):
                negs[pattern.pattern].append(row["score"])
                break
        else:  # this else is for the for loop: ``for pattern in regexps:``
            raise ValueError(
                f"No regexps: {regexps} match `{row[regexp_column]}' "
                f"from `{regexp_column}' column."
            )
    logger.debug(f"Found {len(negs)} different PAIs names: {list(negs.keys())}")
    return pos, negs


def split_csv_pad(filename):
    """Loads PAD scores from a CSV score file, splits them by attack vs
    bonafide.

    The CSV must contain a ``is_bonafide`` column with each field either
    ``True`` or ``False`` (case insensitive).

    Parameters
    ----------
    filename: str
        The path to a CSV file containing all the scores.

    Returns
    -------
    tuple
        Tuple of 1D-arrays: (attack, bonafide). The negative (attacks) and
        positives (bonafide) scores.
    """
    logger.debug(f"Loading CSV score file: '{filename}'")
    split_scores = defaultdict(list)
    for row in _iterate_csv_score_file(filename):
        if row["is_bonafide"].lower() == "true":
            split_scores["bonafide"].append(row["score"])
        else:
            split_scores["attack"].append(row["score"])
    logger.debug(
        f"Found {len(split_scores['attack'])} negative (attack), and"
        f"{len(split_scores['bonafide'])} positive (bonafide) scores."
    )
    # Cast the scores to numpy float
    for key, scores in split_scores.items():
        split_scores[key] = numpy.array(scores, dtype=numpy.float64)
    return split_scores["attack"], split_scores["bonafide"]



"""Prints Cross-db metrics analysis
"""
import itertools
import json
import logging
import math
import os

import click
import jinja2
import yaml

from clapper.click import log_parameters, verbosity_option
from tabulate import tabulate

from bob.bio.base.score.load import get_negatives_positives, load_score
from bob.measure import farfrr
from bob.measure.script import common_options
from bob.measure.utils import get_fta

from ..error_utils import calc_threshold
from .pad_commands import CRITERIA

logger = logging.getLogger(__name__)


def bool_option(name, short_name, desc, dflt=False, **kwargs):
    """Generic provider for boolean options

    Parameters
    ----------
    name : str
        name of the option
    short_name : str
        short name for the option
    desc : str
        short description for the option
    dflt : bool or None
        Default value
    **kwargs
        All kwargs are passed to click.option.

    Returns
    -------
    ``callable``
        A decorator to be used for adding this option.
    """

    def custom_bool_option(func):
        def callback(ctx, param, value):
            ctx.meta[name.replace("-", "_")] = value
            return value

        return click.option(
            "-%s/-n%s" % (short_name, short_name),
            "--%s/--no-%s" % (name, name),
            default=dflt,
            help=desc,
            show_default=True,
            callback=callback,
            is_eager=True,
            **kwargs,
        )(func)

    return custom_bool_option


def _ordered_load(stream, Loader=yaml.Loader, object_pairs_hook=dict):
    """Loads the contents of the YAML stream into :py:class:`collections.OrderedDict`'s

    See: https://stackoverflow.com/questions/5121931/in-python-how-can-you-load-yaml-mappings-as-ordereddicts

    """

    class OrderedLoader(Loader):
        pass

    def construct_mapping(loader, node):
        loader.flatten_mapping(node)
        return object_pairs_hook(loader.construct_pairs(node))

    OrderedLoader.add_constructor(
        yaml.resolver.BaseResolver.DEFAULT_MAPPING_TAG, construct_mapping
    )

    return yaml.load(stream, OrderedLoader)


def expand(data):
    """Generates configuration sets based on the YAML input contents

    For an introduction to the YAML mark-up, just search the net. Here is one of
    its references: https://en.wikipedia.org/wiki/YAML

    A configuration set corresponds to settings for **all** variables in the
    input template that needs replacing. For example, if your template mentions
    the variables ``name`` and ``version``, then each configuration set should
    yield values for both ``name`` and ``version``.

    For example:

    .. code-block:: yaml

       name: [john, lisa]
       version: [v1, v2]


    This should yield to the following configuration sets:

    .. code-block:: python

       [
         {'name': 'john', 'version': 'v1'},
         {'name': 'john', 'version': 'v2'},
         {'name': 'lisa', 'version': 'v1'},
         {'name': 'lisa', 'version': 'v2'},
       ]


    Each key in the input file should correspond to either an object or a YAML
    array. If the object is a list, then we'll iterate over it for every possible
    combination of elements in the lists. If the element in question is not a
    list, then it is considered unique and repeated for each yielded
    configuration set. Example

    .. code-block:: yaml

       name: [john, lisa]
       version: [v1, v2]
       text: >
          hello,
          world!

    Should yield to the following configuration sets:

    .. code-block:: python

       [
         {'name': 'john', 'version': 'v1', 'text': 'hello, world!'},
         {'name': 'john', 'version': 'v2', 'text': 'hello, world!'},
         {'name': 'lisa', 'version': 'v1', 'text': 'hello, world!'},
         {'name': 'lisa', 'version': 'v2', 'text': 'hello, world!'},
       ]

    Keys starting with one `_` (underscore) are treated as "unique" objects as
    well. Example:

    .. code-block:: yaml

       name: [john, lisa]
       version: [v1, v2]
       _unique: [i1, i2]

    Should yield to the following configuration sets:

    .. code-block:: python

       [
         {'name': 'john', 'version': 'v1', '_unique': ['i1', 'i2']},
         {'name': 'john', 'version': 'v2', '_unique': ['i1', 'i2']},
         {'name': 'lisa', 'version': 'v1', '_unique': ['i1', 'i2']},
         {'name': 'lisa', 'version': 'v2', '_unique': ['i1', 'i2']},
       ]


    Parameters:

      data (str): YAML data to be parsed


    Yields:

      dict: A dictionary of key-value pairs for building the templates

    """

    data = _ordered_load(data, yaml.SafeLoader)

    # separates "unique" objects from the ones we have to iterate
    # pre-assemble return dictionary
    iterables = dict()
    unique = dict()
    for key, value in data.items():
        if isinstance(value, list) and not key.startswith("_"):
            iterables[key] = value
        else:
            unique[key] = value

    # generates all possible combinations of iterables
    for values in itertools.product(*iterables.values()):
        retval = dict(unique)
        keys = list(iterables.keys())
        retval.update(dict(zip(keys, values)))
        yield retval


@click.command(
    epilog="""\b
Examples:
  $ bob pad cross 'results/{{ evaluation.database }}/{{ algorithm }}/{{ evaluation.protocol }}/scores/scores-{{ group }}' \
    -td replaymobile \
    -d replaymobile -p grandtest \
    -d oulunpu -p Protocol_1 \
    -a replaymobile_grandtest_frame-diff-svm \
    -a replaymobile_grandtest_qm-svm-64 \
    -a replaymobile_grandtest_lbp-svm-64 \
    > replaymobile.rst &
"""
)
@click.argument("score_jinja_template")
@click.option(
    "-d",
    "--database",
    "databases",
    multiple=True,
    required=True,
    show_default=True,
    help="Names of the evaluation databases",
)
@click.option(
    "-p",
    "--protocol",
    "protocols",
    multiple=True,
    required=True,
    show_default=True,
    help="Names of the protocols of the evaluation databases",
)
@click.option(
    "-a",
    "--algorithm",
    "algorithms",
    multiple=True,
    required=True,
    show_default=True,
    help="Names of the algorithms",
)
@click.option(
    "-n",
    "--names",
    type=click.File("r"),
    help="Name of algorithms to show in the table. Provide a path "
    "to a json file maps algorithm names to names that you want to "
    "see in the table.",
)
@click.option(
    "-td",
    "--train-database",
    required=True,
    help="The database that was used to train the algorithms.",
)
@click.option(
    "-pn",
    "--pai-names",
    type=click.File("r"),
    help="Name of PAIs to compute the errors per PAI. Provide a path "
    "to a json file maps attack_type in scores to PAIs that you want to "
    "see in the table.",
)
@click.option(
    "-g",
    "--group",
    "groups",
    multiple=True,
    show_default=True,
    default=["train", "dev", "eval"],
)
@bool_option("sort", "s", "whether the table should be sorted.", True)
@common_options.criterion_option(lcriteria=CRITERIA, check=False)
@common_options.far_option()
@common_options.table_option()
@common_options.output_log_metric_option()
@common_options.decimal_option(dflt=2, short="-dec")
@verbosity_option(logger)
@click.pass_context
def cross(
    ctx,
    score_jinja_template,
    databases,
    protocols,
    algorithms,
    names,
    train_database,
    pai_names,
    groups,
    sort,
    decimal,
    verbose,
    **kwargs,
):
    """Cross-db analysis metrics"""
    log_parameters(logger)

    names = {} if names is None else json.load(names)

    env = jinja2.Environment(undefined=jinja2.StrictUndefined)

    data = {
        "evaluation": [
            {"database": db, "protocol": proto}
            for db, proto in zip(databases, protocols)
        ],
        "algorithm": algorithms,
        "group": groups,
    }

    metrics = {}

    for variables in expand(yaml.dump(data, Dumper=yaml.SafeDumper)):
        logger.debug(variables)

        score_path = env.from_string(score_jinja_template).render(variables)
        logger.info(score_path)

        database, protocol, algorithm, group = (
            variables["evaluation"]["database"],
            variables["evaluation"]["protocol"],
            variables["algorithm"],
            variables["group"],
        )

        # if algorithm name does not have train_database name in it.
        if train_database not in algorithm and database != train_database:
            score_path = score_path.replace(
                algorithm, database + "_" + algorithm
            )
            logger.info("Score path changed to: %s", score_path)

        if not os.path.exists(score_path):
            metrics[(database, protocol, algorithm, group)] = (
                float("nan"),
            ) * 5
            continue

        scores = load_score(score_path)
        neg, pos = get_negatives_positives(scores)
        (neg, pos), fta = get_fta((neg, pos))

        if group == "eval":
            threshold = metrics[(database, protocol, algorithm, "dev")][1]
        else:
            try:
                threshold = calc_threshold(
                    ctx.meta["criterion"],
                    pos,
                    [neg],
                    neg,
                    ctx.meta["far_value"],
                )
            except RuntimeError:
                logger.error("Something wrong with {}".format(score_path))
                raise

        far, frr = farfrr(neg, pos, threshold)
        hter = (far + frr) / 2

        metrics[(database, protocol, algorithm, group)] = (
            hter,
            threshold,
            fta,
            far,
            frr,
        )

    logger.debug("metrics: %s", metrics)

    headers = ["Algorithms"]
    for db in databases:
        headers += [db + "\nEER_t", "\nEER_d", "\nAPCER", "\nBPCER", "\nACER"]
    rows = []

    # sort the algorithms based on HTER test, EER dev, EER train
    train_protocol = protocols[databases.index(train_database)]
    if sort:

        def sort_key(alg):
            r = []
            for grp in ("eval", "dev", "train"):
                hter = metrics[(train_database, train_protocol, alg, group)][0]
                r.append(1 if math.isnan(hter) else hter)
            return tuple(r)

        algorithms = sorted(algorithms, key=sort_key)

    for algorithm in algorithms:
        name = algorithm.replace(train_database + "_", "")
        name = name.replace(train_protocol + "_", "")
        name = names.get(name, name)
        rows.append([name])
        for database, protocol in zip(databases, protocols):
            cell = []
            for group in groups:
                hter, threshold, fta, far, frr = metrics[
                    (database, protocol, algorithm, group)
                ]
                if group == "eval":
                    cell += [far, frr, hter]
                else:
                    cell += [hter]
            cell = [round(c * 100, decimal) for c in cell]
            rows[-1].extend(cell)

    title = " Trained on {} ".format(train_database)
    title_line = "\n" + "=" * len(title) + "\n"
    # open log file for writing if any
    ctx.meta["log"] = (
        ctx.meta["log"]
        if ctx.meta["log"] is None
        else open(ctx.meta["log"], "w")
    )
    click.echo(title_line + title + title_line, file=ctx.meta["log"])
    click.echo(
        tabulate(rows, headers, ctx.meta["tablefmt"], floatfmt=".1f"),
        file=ctx.meta["log"],
    )



#!/usr/bin/env python
# vim: set fileencoding=utf-8 :

import os

import pkg_resources

# -- General configuration -----------------------------------------------------

# If your documentation needs a minimal Sphinx version, state it here.
needs_sphinx = "1.3"

# Add any Sphinx extension module names here, as strings. They can be extensions
# coming with Sphinx (named 'sphinx.ext.*') or your custom ones.
extensions = [
    "sphinx.ext.todo",
    "sphinx.ext.coverage",
    "sphinx.ext.ifconfig",
    "sphinx.ext.autodoc",
    "sphinx.ext.autosummary",
    "sphinx.ext.doctest",
    "sphinx.ext.graphviz",
    "sphinx.ext.intersphinx",
    "sphinx.ext.napoleon",
    "sphinx.ext.viewcode",
    "sphinx.ext.mathjax",
    "auto_intersphinx",
]

# Be picky about warnings
nitpicky = True

# Ignores stuff we can't easily resolve on other project's sphinx manuals
nitpick_ignore = []
keep_warnings = True

# Allows the user to override warnings from a separate file
if os.path.exists("nitpick-exceptions.txt"):
    for line in open("nitpick-exceptions.txt"):
        if line.strip() == "" or line.startswith("#"):
            continue
        dtype, target = line.split(None, 1)
        target = target.strip()
        try:  # python 2.x
            target = unicode(target)
        except NameError:
            pass
        nitpick_ignore.append((dtype, target))

# Always includes todos
todo_include_todos = True

# Generates auto-summary automatically
autosummary_generate = True

# Create numbers on figures with captions
numfig = True

# If we are on OSX, the 'dvipng' path maybe different
dvipng_osx = "/opt/local/libexec/texlive/binaries/dvipng"
if os.path.exists(dvipng_osx):
    pngmath_dvipng = dvipng_osx

# Add any paths that contain templates here, relative to this directory.
templates_path = ["_templates"]

# The suffix of source filenames.
source_suffix = ".rst"

# The encoding of source files.
# source_encoding = 'utf-8-sig'

# The master toctree document.
master_doc = "index"

# General information about the project.
project = "bob.pad.base"
import time

copyright = "%s, Idiap Research Institute" % time.strftime("%Y")

# Grab the setup entry
distribution = pkg_resources.require(project)[0]

# The version info for the project you're documenting, acts as replacement for
# |version| and |release|, also used in various other places throughout the
# built documents.
#
# The short X.Y version.
version = distribution.version
# The full version, including alpha/beta/rc tags.
release = distribution.version

# The language for content autogenerated by Sphinx. Refer to documentation
# for a list of supported languages.
# language = None

# There are two options for replacing |today|: either, you set today to some
# non-false value, then it is used:
# today = ''
# Else, today_fmt is used as the format for a strftime call.
# today_fmt = '%B %d, %Y'

# List of patterns, relative to source directory, that match files and
# directories to ignore when looking for source files.
exclude_patterns = ["links.rst"]

# The reST default role (used for this markup: `text`) to use for all documents.
# default_role = None

# If true, '()' will be appended to :func: etc. cross-reference text.
# add_function_parentheses = True

# If true, the current module name will be prepended to all description
# unit titles (such as .. function::).
# add_module_names = True

# If true, sectionauthor and moduleauthor directives will be shown in the
# output. They are ignored by default.
# show_authors = False

# The name of the Pygments (syntax highlighting) style to use.
pygments_style = "sphinx"

# A list of ignored prefixes for module index sorting.
# modindex_common_prefix = []

# Some variables which are useful for generated material
project_variable = project.replace(".", "_")
short_description = (
    "Bob framework for running presentation attack detection experiments"
)
owner = ["Idiap Research Institute"]


# -- Options for HTML output ---------------------------------------------------

# The theme to use for HTML and HTML Help pages.  See the documentation for
# a list of builtin themes.
import sphinx_rtd_theme

html_theme = "sphinx_rtd_theme"

# Theme options are theme-specific and customize the look and feel of a theme
# further.  For a list of options available for each theme, see the
# documentation.
# html_theme_options = {}

# Add any paths that contain custom themes here, relative to this directory.
html_theme_path = [sphinx_rtd_theme.get_html_theme_path()]

# The name for this set of Sphinx documents.  If None, it defaults to
# "<project> v<release> documentation".
# html_title = None

# A shorter title for the navigation bar.  Default is the same as html_title.
# html_short_title = project_variable

# The name of an image file (relative to this directory) to place at the top
# of the sidebar.
html_logo = "img/logo.png"

# The name of an image file (within the static path) to use as favicon of the
# docs.  This file should be a Windows icon file (.ico) being 16x16 or 32x32
# pixels large.
html_favicon = "img/favicon.ico"

# Add any paths that contain custom static files (such as style sheets) here,
# relative to this directory. They are copied after the builtin static files,
# so a file named "default.css" will overwrite the builtin "default.css".
# html_static_path = ['_static']

# If not '', a 'Last updated on:' timestamp is inserted at every page bottom,
# using the given strftime format.
# html_last_updated_fmt = '%b %d, %Y'

# If true, SmartyPants will be used to convert quotes and dashes to
# typographically correct entities.
# html_use_smartypants = True

# Custom sidebar templates, maps document names to template names.
# html_sidebars = {}

# Additional templates that should be rendered to pages, maps page names to
# template names.
# html_additional_pages = {}

# If false, no module index is generated.
# html_domain_indices = True

# If false, no index is generated.
# html_use_index = True

# If true, the index is split into individual pages for each letter.
# html_split_index = False

# If true, links to the reST sources are added to the pages.
# html_show_sourcelink = True

# If true, "Created using Sphinx" is shown in the HTML footer. Default is True.
# html_show_sphinx = True

# If true, "(C) Copyright ..." is shown in the HTML footer. Default is True.
# html_show_copyright = True

# If true, an OpenSearch description file will be output, and all pages will
# contain a <link> tag referring to it.  The value of this option must be the
# base URL from which the finished HTML is served.
# html_use_opensearch = ''

# This is the file name suffix for HTML files (e.g. ".xhtml").
# html_file_suffix = None

# Output file base name for HTML help builder.
htmlhelp_basename = project_variable + "_doc"


# -- Post configuration --------------------------------------------------------

# Included after all input documents
rst_epilog = """
.. |project| replace:: Bob
.. |version| replace:: %s
.. |current-year| date:: %%Y
""" % (
    version,
)

# Default processing flags for sphinx
autoclass_content = "class"
autodoc_member_order = "bysource"
autodoc_default_options = {
    "members": True,
    "undoc-members": True,
    "show-inheritance": True,
}

auto_intersphinx_packages = [
    ("python", "3"),
    "numpy",
    "bob.extension",
    "bob.pipelines",
    "bob.bio.base",
    "bob.pad.face",
    "setuptools",
    "dask",
    "scikit-learn",
]
auto_intersphinx_catalog = "catalog.json"



"""The main entry for bob.pad.
"""
import click
import pkg_resources

from clapper.click import AliasedGroup
from click_plugins import with_plugins


@with_plugins(pkg_resources.iter_entry_points("bob.pad.cli"))
@click.group(cls=AliasedGroup)
def pad():
    """Presentation Attack Detection related commands."""
    pass



from .csv_dataset import FileListPadDatabase


# gets sphinx autodoc done right - don't remove it
def __appropriate__(*args):
    """Says object was actually declared here, and not in the import module.
    Fixing sphinx warnings of not being able to find classes, when path is shortened.
    Parameters:

      *args: An iterable of objects to modify

    Resolves `Sphinx referencing issues
    <https://github.com/sphinx-doc/sphinx/issues/3048>`
    """

    for obj in args:
        obj.__module__ = __name__


__appropriate__(
    FileListPadDatabase,
)
__all__ = [_ for _ in dir() if not _.startswith("_")]



"""The main entry for bob pad commands.
"""
import logging
import os

from csv import DictWriter
from functools import partial

import click
import numpy

from clapper.click import verbosity_option

import bob.measure.script.figure as measure_figure

from bob.measure.script import common_options

from ..error_utils import split_csv_pad, split_csv_pad_per_pai
from . import pad_figure as figure

logger = logging.getLogger(__name__)

SCORE_FORMAT = "Files must be in CSV format."
CRITERIA = (
    "eer",
    "min-hter",
    "far",
    "bpcer5000",
    "bpcer2000",
    "bpcer1000",
    "bpcer500",
    "bpcer200",
    "bpcer100",
    "bpcer50",
    "bpcer20",
    "bpcer10",
    "bpcer5",
    "bpcer2",
    "bpcer1",
    "apcer5000",
    "apcer2000",
    "apcer1000",
    "apcer500",
    "apcer200",
    "apcer100",
    "apcer50",
    "apcer20",
    "apcer10",
    "apcer5",
    "apcer2",
    "apcer1",
)


def metrics_option(
    sname="-m",
    lname="--metrics",
    name="metrics",
    help="List of metrics to print. Provide a string with comma separated metric "
    "names. For possible values see the default value.",
    default="apcer_pais,apcer_ap,bpcer,acer,fta,fpr,fnr,hter,far,frr,precision,recall,f1_score,auc,auc-log-scale",
    **kwargs,
):
    """The metrics option"""

    def custom_metrics_option(func):
        def callback(ctx, param, value):
            if value is not None:
                value = value.split(",")
            ctx.meta[name] = value
            return value

        return click.option(
            sname,
            lname,
            default=default,
            help=help,
            show_default=True,
            callback=callback,
            **kwargs,
        )(func)

    return custom_metrics_option


def regexps_option(
    help="A list of regular expressions (by repeating this option) to be used to "
    "categorize PAIs. Each regexp must match one type of PAI.",
    **kwargs,
):
    def custom_regexps_option(func):
        def callback(ctx, param, value):
            ctx.meta["regexps"] = value
            return value

        return click.option(
            "-r",
            "--regexps",
            default=None,
            multiple=True,
            help=help,
            callback=callback,
            **kwargs,
        )(func)

    return custom_regexps_option


def regexp_column_option(
    help="The column in the score files to match the regular expressions against.",
    **kwargs,
):
    def custom_regexp_column_option(func):
        def callback(ctx, param, value):
            ctx.meta["regexp_column"] = value
            return value

        return click.option(
            "-rc",
            "--regexp-column",
            default="attack_type",
            help=help,
            show_default=True,
            callback=callback,
            **kwargs,
        )(func)

    return custom_regexp_column_option


def gen_pad_csv_scores(
    filename, mean_match, mean_attacks, n_attack_types, n_clients, n_samples
):
    """Generates a CSV file containing random scores for PAD."""
    columns = [
        "claimed_id",
        "test_label",
        "is_bonafide",
        "attack_type",
        "sample_n",
        "score",
    ]
    with open(filename, "w") as f:
        writer = DictWriter(f, fieldnames=columns)
        writer.writeheader()
        # Bonafide rows
        for client_id in range(n_clients):
            for sample in range(n_samples):
                writer.writerow(
                    {
                        "claimed_id": client_id,
                        "test_label": f"client/real/{client_id:03d}",
                        "is_bonafide": "True",
                        "attack_type": None,
                        "sample_n": sample,
                        "score": numpy.random.normal(loc=mean_match),
                    }
                )
        # Attacks rows
        for attack_type in range(n_attack_types):
            for client_id in range(n_clients):
                for sample in range(n_samples):
                    writer.writerow(
                        {
                            "claimed_id": client_id,
                            "test_label": f"client/attack/{client_id:03d}",
                            "is_bonafide": "False",
                            "attack_type": f"type_{attack_type}",
                            "sample_n": sample,
                            "score": numpy.random.normal(
                                loc=mean_attacks[
                                    attack_type % len(mean_attacks)
                                ]
                            ),
                        }
                    )


@click.command()
@click.argument("outdir")
@click.option(
    "-mm", "--mean-match", default=10, type=click.FLOAT, show_default=True
)
@click.option(
    "-ma",
    "--mean-attacks",
    default=[-10, -6],
    type=click.FLOAT,
    show_default=True,
    multiple=True,
)
@click.option(
    "-c", "--n-clients", default=10, type=click.INT, show_default=True
)
@click.option("-s", "--n-samples", default=2, type=click.INT, show_default=True)
@click.option("-a", "--n-attacks", default=2, type=click.INT, show_default=True)
@verbosity_option(logger)
@click.pass_context
def gen(
    ctx,
    outdir,
    mean_match,
    mean_attacks,
    n_clients,
    n_samples,
    n_attacks,
    **kwargs,
):
    """Generate random scores.
    Generates random scores in CSV format. The scores are generated
    using Gaussian distribution whose mean is an input
    parameter. The generated scores can be used as hypothetical datasets.
    n-attacks defines the number of different type of attacks generated (like print and
    mask). When multiples attacks are present, the mean-attacks option can be set
    multiple times, specifying the mean of each attack scores distribution.

    Example:

    bob pad gen results/generated/scores-dev.csv -a 3 -ma 2 -ma 5 -ma 7 -mm 8
    """
    numpy.random.seed(0)
    gen_pad_csv_scores(
        os.path.join(outdir, "scores-dev.csv"),
        mean_match,
        mean_attacks,
        n_attacks,
        n_clients,
        n_samples,
    )
    gen_pad_csv_scores(
        os.path.join(outdir, "scores-eval.csv"),
        mean_match,
        mean_attacks,
        n_attacks,
        n_clients,
        n_samples,
    )


@common_options.metrics_command(
    common_options.METRICS_HELP.format(
        names="FtA, APCER_AP, BPCER, FPR, FNR, FAR, FRR, ACER, HTER, precision, recall, f1_score",
        criteria=CRITERIA,
        score_format=SCORE_FORMAT,
        hter_note="Note that APCER_AP = max(APCER_pais), BPCER=FNR, "
        "FAR = FPR * (1 - FtA), "
        "FRR = FtA + FNR * (1 - FtA), "
        "ACER = (APCER_AP + BPCER) / 2, "
        "and HTER = (FPR + FNR) / 2. "
        "You can control which metrics are printed using the --metrics option. "
        "You can use --regexps and --regexp_column options to change the behavior "
        "of finding Presentation Attack Instrument (PAI) types",
        command="bob pad metrics",
    ),
    criteria=CRITERIA,
    check_criteria=False,
    epilog="""\b
More Examples:
\b
bob pad metrics -vvv -e -lg IQM,LBP -r print -r video -m fta,apcer_pais,apcer_ap,bpcer,acer,hter \
/scores/oulunpu/{qm-svm,lbp-svm}/Protocol_1/scores/scores-{dev,eval}

See also ``bob pad multi-metrics``.
""",
)
@regexps_option()
@regexp_column_option()
@metrics_option()
def metrics(ctx, scores, evaluation, regexps, regexp_column, metrics, **kwargs):
    load_fn = partial(
        split_csv_pad_per_pai, regexps=regexps, regexp_column=regexp_column
    )
    process = figure.Metrics(ctx, scores, evaluation, load_fn, metrics)
    process.run()


@common_options.roc_command(
    common_options.ROC_HELP.format(
        score_format=SCORE_FORMAT, command="bob pad roc"
    )
)
def roc(ctx, scores, evaluation, **kwargs):
    process = figure.Roc(ctx, scores, evaluation, split_csv_pad)
    process.run()


@common_options.det_command(
    common_options.DET_HELP.format(
        score_format=SCORE_FORMAT, command="bob pad det"
    )
)
def det(ctx, scores, evaluation, **kwargs):
    process = figure.Det(ctx, scores, evaluation, split_csv_pad)
    process.run()


@common_options.epc_command(
    common_options.EPC_HELP.format(
        score_format=SCORE_FORMAT, command="bob pad epc"
    )
)
def epc(ctx, scores, **kwargs):
    process = measure_figure.Epc(ctx, scores, True, split_csv_pad, hter="ACER")
    process.run()


@common_options.hist_command(
    common_options.HIST_HELP.format(
        score_format=SCORE_FORMAT, command="bob pad hist"
    )
)
def hist(ctx, scores, evaluation, **kwargs):
    process = figure.Hist(ctx, scores, evaluation, split_csv_pad)
    process.run()


@common_options.evaluate_command(
    common_options.EVALUATE_HELP.format(
        score_format=SCORE_FORMAT, command="bob pad evaluate"
    ),
    criteria=CRITERIA,
)
def evaluate(ctx, scores, evaluation, **kwargs):
    common_options.evaluate_flow(
        ctx, scores, evaluation, metrics, roc, det, epc, hist, **kwargs
    )


@common_options.multi_metrics_command(
    common_options.MULTI_METRICS_HELP.format(
        names="FtA, APCER, BPCER, FAR, FRR, ACER, HTER, precision, recall, f1_score",
        criteria=CRITERIA,
        score_format=SCORE_FORMAT,
        command="bob pad multi-metrics",
    ),
    criteria=CRITERIA,
    epilog="""\b
More examples:

\b
bob pad multi-metrics -vvv -e -pn 6 -lg IQM,LBP -r print -r video \
/scores/oulunpu/{qm-svm,lbp-svm}/Protocol_3_{1,2,3,4,5,6}/scores/scores-{dev,eval}

See also ``bob pad metrics``.
""",
)
@regexps_option()
@regexp_column_option()
@metrics_option(default="fta,apcer_pais,apcer_ap,bpcer,acer,hter")
def multi_metrics(
    ctx,
    scores,
    evaluation,
    protocols_number,
    regexps,
    regexp_column,
    metrics,
    **kwargs,
):
    ctx.meta["min_arg"] = protocols_number * (2 if evaluation else 1)
    load_fn = partial(
        split_csv_pad_per_pai, regexps=regexps, regexp_column=regexp_column
    )
    process = figure.MultiMetrics(ctx, scores, evaluation, load_fn, metrics)
    process.run()



# isort: skip_file
from . import database  # noqa: F401
from . import error_utils  # noqa: F401
from . import script  # noqa: F401

# gets sphinx autodoc done right - don't remove it
__all__ = [_ for _ in dir() if not _.startswith("_")]
